"use strict";

let EXPORTED_SYMBOLS = ['WinInt'];

const Cc = Components.classes;
const Ci = Components.interfaces;

let mds = {};
let {ctypes} = Components.utils.import('resource://gre/modules/ctypes.jsm', mds)
let {Sync} = Components.utils.import('resource://mpint/Sync.js', mds);

const WM_COPYDATA = 74;
const PROCESS_QUERY_INFORMATION = 0x0400;
const PROCESS_VM_READ = 0x0010;
const CP_ACP = 0;

const DWORD = ctypes.uint32_t;
const UINT = ctypes.unsigned_int;
const BOOL = ctypes.bool;
const HANDLE = ctypes.size_t;
const HWND = HANDLE;
const HMODULE = HANDLE;
const WPARAM = ctypes.size_t;
const LPARAM = ctypes.size_t;
const LRESULT = ctypes.size_t;
const ULONG_PTR = ctypes.uintptr_t;
const PVOID = ctypes.voidptr_t;
const LPCTSTR = ctypes.jschar.ptr;
const LPCWSTR = ctypes.jschar.ptr;
const LPTSTR = ctypes.jschar.ptr;
const LPSTR = ctypes.char.ptr;
const LPCSTR = ctypes.char.ptr;

const COPYDATASTRUCT = new ctypes.StructType('COPYDATASTRUCT',
  [
    {'dwData': ULONG_PTR},
    {'cbData': DWORD},
    {'lpData': PVOID}
  ]);

const OSVersionInfo = new ctypes.StructType('OSVERSIONINFO',
  [
    {'dwOSVersionInfoSize': DWORD},
    {'dwMajorVersion': DWORD},
    {'dwMinorVersion': DWORD},
    {'dwBuildNumber': DWORD},
    {'dwPlatformId': DWORD},
    {'szCSDVersion': new ctypes.ArrayType(ctypes.jschar, 128)}
  ]);

var CallBackABI;
var WinABI;

if (ctypes.size_t.size == 8) {
  CallBackABI = ctypes.default_abi;
  WinABI = ctypes.default_abi;
} else {
  CallBackABI = ctypes.stdcall_abi;
  WinABI = ctypes.winapi_abi;
}

const EnumWindowsProc = ctypes.FunctionType(
  CallBackABI, BOOL, [HWND, LPARAM]);

const RegKeyClass = '@mozilla.org/windows-registry-key;1';
const WRegKey = Ci.nsIWindowsRegKey;
const HKCR = 0;
const HKCU = 1;
const HKLM = 2;

const AppPaths = 'Software\\Microsoft\\Windows\\CurrentVersion\\App Paths\\';
const MDProp = 'Software\\MetaProducts\\Mass Downloader';
const DEProp = 'Software\\MetaProducts\\Download Express';

function getRegKey(reg, key) {
  switch (key) {
    case HKCR:
      return reg.ROOT_KEY_CLASSES_ROOT;
    case HKCU:
      return reg.ROOT_KEY_CURRENT_USER;
    case HKLM:
      return reg.ROOT_KEY_LOCAL_MACHINE;
  }
};

function getRegString(key, path, name) {
  var res = '';
  try {
    let reg = Cc[RegKeyClass].createInstance(WRegKey);
    reg.open(getRegKey(reg, key), path, WRegKey.ACCESS_READ);
    try {
      res = reg.readStringValue(name);
    } catch(ex) {}
    reg.close;
  } catch(ex) {}
  return res;
};

function FileExists(fname) {
  var f = Cc['@mozilla.org/file/local;1'].createInstance(Ci.nsILocalFile);
  f.initWithPath(fname);
  return f.exists();
}

function RunIt(exepath, args) {
  try {
    var f = Cc['@mozilla.org/file/local;1'].createInstance(Ci.nsILocalFile);
    var pr = Cc['@mozilla.org/process/util;1'].createInstance(Ci.nsIProcess);
    f.initWithPath(exepath);
    if (f.exists()) {
      pr.init(f);
      pr.runw(false, args, args.length);
    }
  } catch(ex) {}
}

function makeExpiration(offset) {
  var today = new Date();
  var expire = (new Date(today.getTime() + offset * 1000)).toUTCString();
  return expire.substr(0, 7) + '-' + expire.substr(8, 3) + '-' + expire.substr(12);
}

function Data() {
  var user32dll;
  var kernel32dll;
  var wininetdll;
  var psapidll;

  var FindWindow;
  var EnumWindows;
  var SendMessage;
  var GetClassName;
  var GetWindowThreadProcessId;
  var OpenProcess;
  var CloseHandle;
  var WideCharToMultiByte;
  var GetVersionEx;
  var InternetSetCookie;
  var GetModuleBaseName;

  this.Loaddlls = function() {
    user32dll = ctypes.open('user32.dll');
    kernel32dll = ctypes.open('kernel32.dll');
    wininetdll = ctypes.open('wininet.dll');
    psapidll = ctypes.open('psapi.dll');

    FindWindow = user32dll.declare('FindWindowW',
      WinABI, HWND, LPCTSTR, LPCTSTR);

    EnumWindows = user32dll.declare('EnumWindows',
      WinABI, BOOL, EnumWindowsProc.ptr, LPARAM);

    SendMessage = user32dll.declare('SendMessageW',
      WinABI, LRESULT, HWND, UINT, WPARAM, LPARAM);

    GetClassName = user32dll.declare('GetClassNameW',
      WinABI, ctypes.int, HWND, LPTSTR, ctypes.int);

    GetWindowThreadProcessId = user32dll.declare('GetWindowThreadProcessId',
      WinABI, DWORD, HWND, DWORD.ptr);

    OpenProcess = kernel32dll.declare('OpenProcess',
      WinABI, HANDLE, DWORD, BOOL, DWORD);

    CloseHandle = kernel32dll.declare('CloseHandle',
      WinABI, BOOL, HANDLE);

    WideCharToMultiByte = kernel32dll.declare('WideCharToMultiByte',
      WinABI, ctypes.int,
      UINT, DWORD, LPCWSTR, ctypes.int, LPSTR, ctypes.int, LPCSTR, BOOL.ptr);
    
    GetVersionEx = kernel32dll.declare('GetVersionExW',
      WinABI, BOOL, OSVersionInfo.ptr);

    InternetSetCookie = wininetdll.declare('InternetSetCookieW',
      WinABI, BOOL, LPCTSTR, LPCTSTR, LPCTSTR);

    GetModuleBaseName = psapidll.declare('GetModuleBaseNameW',
      WinABI, DWORD, HANDLE, HMODULE, LPTSTR, DWORD);
  }

  this.UnLoaddlls = function() {
    user32dll.close();
    kernel32dll.close();
    wininetdll.close();
    psapidll.close();
  }

  function Wide2Ansi(str) {
    if (str) {
      let i = WideCharToMultiByte(CP_ACP, 0, str, -1, null, 0, null, null);
      let buf = new new ctypes.ArrayType(ctypes.char, i);
      i = WideCharToMultiByte(CP_ACP, 0, str, -1, buf, i, null, null);
      if (i)
        return buf;
    }
    return null;
  }

  function GetCmdLineLength() {
    var OSVI = new OSVersionInfo();
    OSVI.dwOSVersionInfoSize = OSVersionInfo.size;
    GetVersionEx(OSVI.address());
    
    var version = OSVI.dwMajorVersion * 10 + OSVI.dwMinorVersion;
    if (version >= 51)
      return 8191;
    else
      return 2047;
  }
  
  function setCookie(url, cookie) {
    if ((url) && (cookie)) {
      var expire = makeExpiration(300);
      var cookies = cookie.split('; ');
      cookies.forEach(function(s) {
        if (s) {
          let data = s + '; expires = ' + expire;
          InternetSetCookie(url, null, data);
        }
      })
    }
  }

  function GenerateData(url, ref, comment, cookie, fn, mime, IsMD) {
    var cmdLine = 'command=URL=' + url;
    var copyData = 'URL=' + url;
    if (ref) {
      cmdLine += ';Referer=' + ref;
      copyData += '\nReferer=' + ref;
    }
    if (comment) {
      cmdLine += ';Comment=' + comment.replace(/\"/g, '\'');
      copyData += '\nComment=' + comment;
    }
    if (cookie) {
      let s = btoa(cookie);
      cmdLine += ';Cookie=' + s;
      copyData += '\nCookie=' + s;
    }
    if (fn) {
      cmdLine += ';FileName=' + fn;
      copyData += '\nFileName=' + fn;
    }
    if (mime) {
      cmdLine += ';MIMEType=' + mime;
      copyData += '\nMIMEType=' + mime;
    }
    if (IsMD) {
      cmdLine += ';Mode=DEFAULT';
      copyData += '\nMode=DEFAULT';
    }
    return [cmdLine, copyData];
  }

  function GenerateForMulti(url, ref, comment, cookie) {
    var s = 'URL=' + url;
    if (ref)
      s += '\nReferer=' + ref;
    if (comment)
      s += '\nComment=' + comment;
    if (cookie)
      s += '\nCookie=' + btoa(cookie);
    s += '\nMode=DEFAULT';
    return s;
  }
  
  function SendByCopyData(hWnd, command) {
    let cmd = Wide2Ansi(ctypes.jschar.array()(command));
    if (cmd) {
      let CD = new COPYDATASTRUCT();
      CD.lpData = cmd.address();
      CD.cbData = cmd.length - 1;
      CD.dwData = 0;
      SendMessage(hWnd, WM_COPYDATA, 0, ctypes.cast(CD.address(), LPARAM));
    }
  }

  function SendByCommandLine(exe, command) {
    let args = new Array();
    args.push(command);
    RunIt(exe, args);
  }
  
  function GetDEHwnd() {
    var caption = getRegString(HKCU, DEProp, 'ProgramTitle');
    return FindWindow('{B966C3C5-9030-46FC-B930-1E1603701D5C}', caption);
  }

  function GetDEFName() {
    return getRegString(HKCU, DEProp, 'ProgramPath');
  }

  function RunDE() {
    var hWnd = GetDEHwnd();
    if (hWnd == 0) {
      RunIt(GetDEFName(), new Array());
      for (let i = 0; i < 100; i++) {
        Sync.sleep(100);
        hWnd = GetDEHwnd();
        if (hWnd != 0)
          break;
      };
    }
    return hWnd;
  }

  function ProcessFileName(PID) {
    var result = '';
    var handle = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, false, PID);
    if (handle != 0) {
      var buf = new new ctypes.ArrayType(ctypes.jschar, 255);
      if (GetModuleBaseName(handle, 0, buf, 255)) {
        result = buf.readString();
      }
      CloseHandle(handle);
    }
    return result;
  }

  function SearchPD(hwnd, lParam) {
    var result = true;

    var buf = new new ctypes.ArrayType(ctypes.jschar, 255);
    GetClassName(hwnd, buf, 255);
    if (buf.readString() == 'TMainForm') {
      var PID = new DWORD;
      GetWindowThreadProcessId(hwnd, PID.address());
      var PName = ProcessFileName(PID.address().contents);
      if (PName.toLowerCase() == 'pdownloadmanager.exe') {
        LPARAM.ptr(lParam).contents = hwnd;
        result = false;
      }
    }
    return result;
  }

  function GetMDHwnd() {
    var lpCaption = getRegString(HKCU, MDProp, 'MDCaption');
    return FindWindow('TMainForm', lpCaption);
  }

  function GetMDFName() {
    var pd = false;
    var lpRun = getRegString(HKLM, AppPaths + 'massdown.exe', '');
    if (!((lpRun) && (FileExists(lpRun)))) {
      lpRun = getRegString(HKLM, AppPaths + 'pdownloadmanager.exe', '');
      if ((lpRun) && (FileExists(lpRun)))
        pd = true;
      else
        lpRun = '';
    }
    return [lpRun, pd];
  }
  
  function RunMD() {
    var SearchPD_ptr = EnumWindowsProc.ptr(SearchPD);
    var wnd = LPARAM(0);

    var hWnd = GetMDHwnd();
    if (hWnd == 0) {
      let [lpRun, pd] = GetMDFName();
    
      RunIt(lpRun, new Array());
      for (let i = 0; i < 100; i++) {
        Sync.sleep(100);
        hWnd = GetMDHwnd();
        if (hWnd != 0)
          break;
        if (pd) {
          EnumWindows(SearchPD_ptr, ctypes.cast(wnd.address(), LPARAM));
          hWnd = wnd.address().contents;
          if (hWnd != 0)
            break;
        }
      };
    }
    return hWnd;
  }

  this.DownloadWithDE = function(url, ref, comment, cookie, fn, mime) {
    let [cmdLine, copyData] = GenerateData(url, ref, comment, cookie, fn, mime, false);
    let hWnd = GetDEHwnd();
    if (hWnd != 0) {
      setCookie(url, cookie);
      SendByCopyData(hWnd, copyData);
    } else {
      let fname = GetDEFName();
      let len = GetCmdLineLength();
      if ((fname + ' ' + cmdLine).length < len) {
        setCookie(url, cookie);
        SendByCommandLine(fname, cmdLine);
      } else {
        hWnd = RunDE();
        setCookie(url, cookie);
        SendByCopyData(hWnd, copyData);
      }
    }
  }

  this.DownloadWithMD = function(url, ref, comment, cookie, fn, mime) {
    let [cmdLine, copyData] = GenerateData(url, ref, comment, cookie, fn, mime, true);
    let hWnd = GetMDHwnd();
    if (hWnd != 0) {
      setCookie(url, cookie);
      SendByCopyData(hWnd, copyData);
    } else {
      let [fname, ] = GetMDFName();
      let len = GetCmdLineLength();
      if ((fname + ' ' + cmdLine).length < len) {
        setCookie(url, cookie);
        SendByCommandLine(fname, cmdLine);
      } else {
        hWnd = RunMD();
        setCookie(url, cookie);
        SendByCopyData(hWnd, copyData);
      }
    }
  }
  
  this.DownloadURLsWithMD = function(ref, links) {
    let hWnd = RunMD();
    if (hWnd != 0) {
      SendByCopyData(hWnd, GenerateForMulti('Begin.', '', 'Begin.', ''));
      links.forEach(function(link) {
        if (link) {
          setCookie(link.url, link.cookie);
          SendByCopyData(hWnd, GenerateForMulti(link.url, ref, link.comment, link.cookie));
        }
      });
      SendByCopyData(hWnd, GenerateForMulti('End.', '', 'End.', ''));
    }
  }
}

const WinInt = {
  DownloadDEURL: function(url, ref, comment, cookie, fn, mime) {
    if (url) {
      let ct = new Data();
      ct.Loaddlls();
      ct.DownloadWithDE(url, ref, comment, cookie, fn, mime);
      ct.UnLoaddlls();
    }
  },

  DownloadMDURL: function(url, ref, comment, cookie, fn, mime) {
    if (url) {
      let ct = new Data();
      ct.Loaddlls();
      ct.DownloadWithMD(url, ref, comment, cookie, fn, mime);
      ct.UnLoaddlls();
    }
  },

  DownloadMDURLs: function(ref, links) {
    let ct = new Data();
    ct.Loaddlls();
    ct.DownloadURLsWithMD(ref, links);
    ct.UnLoaddlls();
  }
};
